structure KFolyo :> KFolyo =
  (* Hank Dvid kdja alapjn *)
struct

open TFolyo;

(* A bemenetet, ill. a kimenetet a terminlra irnythatjuk Linux alatt
   a /dev/stdin, ill. a /dev/stdout, MS-DOS/Windows alatt mindkt esetben
   a con fjlnv megadsval.
*)

exception KFolyo of string;

(* helyes_feladvany ft = A Folyo.folyo fggvny els paramtere
                         mindenkppen kielgti ezt a prediktumot
*)
fun helyes_feladvany (Meret, MezoL) =
      let
        val (MaxS, MaxO) = Meret
     
        fun helyes_pont (Sor, Oszlop) = 1 <= Sor andalso 1 <= Oszlop andalso
                                     Sor <= MaxS andalso Oszlop <= MaxO
     
        fun helyes_tipus (Kikoto n) = 1 <= n andalso n <= 7
          | helyes_tipus _ = true
     
        fun helyes_mezo (Sor, Oszlop, Tipus) =
              helyes_pont (Sor, Oszlop) andalso helyes_tipus Tipus
     
        fun tavolsag (SOME (S1, O1), SOME (S2, O2)) =
              Math.sqrt(Math.pow(real(S1)-real(S2),2.0)+
              Math.pow(real(O1)-real(O2),2.0))
          | tavolsag _ = 10e6
     
        fun lexikografikusan_hasonlit ((x1,y1,_), (x2,y2,_)) =
              if      x1 < x2 orelse x1 = x2 andalso y1 < y2
              then LESS
              else if x1 > x2 orelse x1 = x2 andalso y1 > y2
              then GREATER
              else EQUAL
     
        fun szomszedok_kulonbozok (rp1::(rps as rp2::_)) =
              lexikografikusan_hasonlit (rp1, rp2) <> EQUAL
              andalso szomszedok_kulonbozok rps
          | szomszedok_kulonbozok _ = true
     
        fun forras (_, _, Forras) = true
          | forras _ = false
     
        fun to (_, _, To) = true
          | to _ = false
     
        fun coords (SOME (S, O, _)) = SOME (S, O)
          | coords NONE = NONE
      in
        (* a tbla mretezse helyes: *)
        1 <= MaxS andalso 1 <= MaxO andalso
     
        (* a MezoL lista minden eleme helyes: *)
        (List.all (fn x => x) (map helyes_mezo MezoL)) andalso
     
        (* a forrs s a t csak a sarkukkal rintkezhet: *)
        tavolsag(coords(List.find forras MezoL),
                 coords(List.find to MezoL)) > 1.0 andalso
                 
        (* nincs kt forrs mez a listban: *)
        List.length (List.filter forras MezoL) < 2 andalso
        
        (* nincs kt t mez a listban: *)
        List.length (List.filter to MezoL) < 2 andalso

        (* a MezoL lista lexikografikusan rendezett: *)
        Listsort.sorted lexikografikusan_hasonlit MezoL andalso

        (* A MezoL lista minden eleme klnbz (Sor,Oszlop)
           koordintra vonatkozik - a rendezettsg miatt elegend
           ezt a szomszdos elemekre kiktni: *)
        szomszedok_kulonbozok MezoL
      end
 
(* folyo_be f = Az f szvegfjlbl beolvasott adatokbl kpzett folytrkp
   folyo_be : string -> TFolyo.folyoterkep
*)
fun folyo_be f =
      let
          fun string2int s =
                let
                    val i = Int.fromString s
                in 
                    if isSome i
                    then valOf i
                    else raise KFolyo("Szintaktikusan hibs adat a(z) " ^
                                      f ^ " fjlban.")
                end

          fun par [a,b] = (a,b)
            | par _ = raise KFolyo("Szintaktikusan hibs adat a(z) " ^
                                   f ^ " fjlban.")

          fun tipus ["folyo"]  = Folyo
            | tipus ["forras"] = Forras
            | tipus ["to"]  = To
            | tipus ["ret"]  = Ret
            | tipus ["kikoto",n] = Kikoto (string2int n)
            | tipus _    = raise KFolyo("Szintaktikusan hibs adat a(z) " ^
                                        f ^ " fjlban.")

          fun mezo (a::b::c) = (string2int a,string2int b,tipus c)
            | mezo _ = raise KFolyo("Szintaktikusan hibs adat a(z) " ^
                                    f ^ " fjlban.")
 
          val tokenizedLine = (String.tokens Char.isSpace) o TextIO.inputLine
 
          fun mezol_be s = if TextIO.endOfStream s
                           then []
                           else (mezo o tokenizedLine) s :: mezol_be s
 
          val istr = TextIO.openIn f

          val meret = (par o (List.map string2int) o tokenizedLine) istr

          val mezol = mezol_be istr

          val terkep = (meret, mezol) before TextIO.closeIn istr
      in
         if helyes_feladvany terkep
         then terkep
         else raise KFolyo("Hibs folytrkp a(z) " ^ f ^ " fjlban.")
      end

(* folyorajz (ft, fv) = a fv kgyvonalnak megfelel folyrajz az ft
                        folytrkpen
   folyorajz : TFolyo.folyoterkep * TFolyo.folyovonal -> string
*)
fun folyorajz (((m,sz), mzk), fv) =
      let
          fun paircmp ((a,b), (c,d)) =
                case Int.compare(a,c) of
                     EQUAL => Int.compare(b,d)
                   | other => other

          fun displayChar Folyo  = #"@"
            | displayChar Forras  = #"F"
            | displayChar To   = #"T"
            | displayChar Ret   = #"R"
            | displayChar (Kikoto n) = chr(ord #"0" + n)

          fun insert ((S,O,c), bmap) = Binarymap.insert(bmap, (S,O), c)

          val pk = Binarymap.mkDict paircmp
          val pk = foldl insert pk (map (fn (S,O) => (S,O,#"@")) fv)
          val pk = foldl insert pk
                         (map (fn (S,O,v) => (S,O,displayChar v)) mzk)

          val tab   = String.concat o List.tabulate
          val vonal = tab (sz, fn _ => "+---") ^ "+\n"

          fun cellarajz S O = case Binarymap.peek(pk, (S+1,O+1)) of
                 SOME c => " " ^ str c ^ " |"
               | NONE   => "   |"

          fun sorrajz i = "|" ^ tab (sz, cellarajz i) ^ "\n" ^ vonal
      in
          vonal ^ tab (m, sorrajz)
      end

(* folyorajzok (ft, fvk) = az fvk megoldslistnak megfelel folyrajzok az
                           ft folytrkpen, res sorokkal elvlasztva
   folyorajzok : TFolyo.folyoterkep * TFolyo.folyovonal list -> string
*)
fun folyorajzok (ft, fvk) =
      foldl (fn (fv,s) => s ^ folyorajz(ft, fv) ^ "\n") "" fvk

(* folyo_ki(m, ft, fvk) = A kt folytrkp fvk megoldslistjt olvashat
                          formban kirja az m szvegfjlba.
   folyo_ki : string * TFolyo.folyoterkep * TFolyo.folyovonal list -> unit
*)
fun folyo_ki(m, ft, fvk) =
      let
          val ostr = TextIO.openOut m
          val _ = TextIO.output(ostr, folyorajzok(ft, fvk))
      in
         TextIO.closeOut ostr
      end

(* megold(f, m) = Beolvas egy feladvnyt az f szvegfjlbl s
                  kirja az sszes megoldst az m szvegfjlba. Ehhez
                  termszetesen szksge van a Folyo.folyo fggvnyre.
                  Eredmnye az f nevet, a megoldsok szmt s a futsi
                  idt mp-ben tartalmaz fzr.
   megold : string * string -> string
*)
fun megold (f, m) =
      let
          val ft = folyo_be f
          val start = Timer.startCPUTimer()
          val fvk = Folyo.folyo ft
          val {usr=tim,...} = Timer.checkCPUTimer start
          val _ = folyo_ki(m, ft, fvk)
      in
         "Feladvny: " ^ f ^ ", " ^
         Int.toString(length fvk) ^ " megolds, " ^
         "id: " ^ Time.fmt 2 tim ^ " mp\n"
      end

end
