% :- type mezo            ---> m(sor, oszlop, mezo_tipus).
% :- type mezo_tipus      ---> folyo ; forras ; to ; ret ; kikoto(dokkszam).
% :- type sor               == integer.
% :- type oszlop            == integer.
% :- type dokkszam          == integer.
% :- type pont            ---> sor-oszlop.
% :- type meret           ---> sor-oszlop.
% :- type folyovonal        == list(pont).

% :- type file              ==  atom.

% :- pred folyo(folyoterkep::in, folyovonal::out).

:- module(folyo_keret,
		  [
		   helyes_feladvany/1,	% helyes_feladvany(folyoterkep::in)
		   folyo_be/2,			% folyo_be(file::in, folyoterkep::out)
		   folyo_ki/3,			% folyo_ki(file::in, folyoterkep::in, folyovonal::in)
		   megold/2,			% megold(file::in, file::in)
		   stopper/2			% stopper(file::in, file::in)
		  ]
		 ).
		 
:- use_module(library(lists), [member/2,append/3,last/2,select/3,non_member/2]).

%-------------- Bejaratok ------------------------------------

% folyo_be(+File, -Feladvany):
% A File-bol beolvashato a Feladvany altal megadott feladat.
folyo_be(File, Feladvany) :-
	open_file(File, read, Stream),
	call_cleanup(read_in(Stream, Characters),
				 close(Stream)),
	parse_data(Characters, Feladvany).

% folyo_ki(+File, +Feladvany, +Folyo):
% A Feladvany feladat Folyo megoldasat kiirja a File allomanyba.
folyo_ki(File, Feladvany, Folyo) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Feladvany, Folyo),
				 (set_output(OOut),close(Stream))).

% megold(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
megold(FileIn, FileOut) :-
	folyo_be(FileIn, Feladvany),
	ensure_loaded(user:folyo),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Feladvany),
				 (set_output(OOut),close(Stream))).

% stopper(+FileIn, +FileOut):
% Mint megold/2, de a futashoz szukseges idot is kiirja.
stopper(FileIn, FileOut) :-
	ensure_loaded(user:folyo),
	statistics(runtime, [T0|_]),
	megold(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Feladat: ~a, futsi id: ~3d s~n', [FileIn,T]).

%-------------- Ellenorzesek ---------------------------------

helyes_feladvany(f(Meret,MezoL)) :-
	% a plya mrete helyes:
	helyes_pont(Meret,inf-inf),
	% nincs nem helyes eleme a MezoL listnak:
	\+ ( member(Mezo, MezoL),
		 \+ helyes_mezo(Mezo, Meret)
	   ),
	% a forrs s a t (ha meg vannak adva) nem lehetnek lszomszdosak,
	% azaz tvolsguk nagyobb 1-nl:
	( member(m(SF,OF,forras),MezoL),
	  member(m(ST,OT,to),MezoL)
	->
	  tavolsaga(SF-OF, ST-OT, Tav),
	  Tav > 1
	; true
	),
	% nincs kt forrs a listban:
	( select(m(_,_,forras),MezoL,MezoLF) -> \+ member(m(_,_,forras),MezoLF) 
	; true
        ),
	% nincs kt t a listban:
	( select(m(_,_,to),MezoL,MezoLT) -> \+ member(m(_,_,to),MezoLT) 
	; true
        ),
	% a MezoL lista lexikografikusan rendezett:
	sort(MezoL, MezoL),
	% a MezoL lista minden eleme klnbz (Sor,Oszlop)
	% koordintra vonatkozik (a rendezettsg miatt elegend ezt a
	% szomszdos elemekre kiktni):
	\+ append(_, [p(S,O,_),p(S,O,_)|_], MezoL).
	
helyes_pont(S-O, MaxS-MaxO) :-
	integer(S), S >=1, S =< MaxS,
	integer(O), O >=1, O =< MaxO.

tavolsaga(S1-O1, S2-O2, T) :-
	T is sqrt( (S1-S2)**2 + (O1-O2)**2 ).
	
helyes_mezo(m(S,O,Tipus),Meret) :-
	helyes_pont(S-O,Meret),
	helyes_tipus(Tipus).
	
helyes_tipus(folyo).
helyes_tipus(forras).
helyes_tipus(to).
helyes_tipus(ret).
helyes_tipus(kikoto(N)) :-
	integer(N), N >= 1, N =<7 .
	
%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).
	
% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.  
% Ha C = -1, akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Table):
% Chars fuzerbol kielemezheto egy Feladvany feladvnyler
parse_data(Chars, Feladvany) :-
	descriptor(Feladvany, Chars, Rest), 
	correct_syntax(Rest),
	helyes_feladvany(Feladvany), !.
parse_data(_, _) :-
	warning('A bemeneti llomny nem megfelel formj.',[]),
	fail.

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :-
	Rest = [_|_], !,
	chars_of_type(mid_line, L, Rest, _),
	warning('Nem vrt karakter(ek): "~s"', [L]),
	fail.
correct_syntax(_).

%-------------- DCG elemzes ----------------------------------

% Kielemezheto egy feladvany-leiro
descriptor(f(R-C,FieldL)) -->
	nat_number(R), nat_number(C),
	newline,
	field_list(FieldL),
	chars_of_type(layout, _).

% Kielemezheto egy (tizes szamrendszerbeli) Int termeszetes szam, amit
% esetleg szokozok eloznek meg.
nat_number(Int) -->
	chars_of_type(horizontal, _),
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

% field_list(FieldL)
% Kielemezheto egy mezo leirasokbol allo FieldL lista.
field_list([m(R,C,Type)|FieldL]) -->
	nat_number(R), nat_number(C), chars_of_type(horizontal, _), field_type(Type),
	newline, !,
	field_list(FieldL).
field_list([]) --> [].

% field_type(Type)
% Kielemezheto a Type mezo tipus leiras.
field_type(folyo) -->		"folyo", !.
field_type(forras) -->		"forras", !.
field_type(to) -->			"to", !.
field_type(ret) -->			"ret", !.
field_type(kikoto(N)) -->	"kikoto", nat_number(N), !.

newline -->
	chars_of_type(horizontal, _),
	char_of_type(vertical, _).

% layout_char(Dir, C):
% horizontal_layout_char(C): C  egy vizszintes iranyu
% nem lathato karakter.
horizontal_layout_char(0' ).
horizontal_layout_char(0'\t).
horizontal_layout_char(0'\r).

% vertical_layout_char(C): C  egy fuggoleges iranyu
% nem lathato karakter.
vertical_layout_char(0'\n).
vertical_layout_char(0'\f).
vertical_layout_char(0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	(	vertical_layout_char(C) -> true
	;	horizontal_layout_char(C)
	).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
char_type(mid_line, C) :-
	\+ vertical_layout_char(C).
char_type(horizontal, C) :-
	horizontal_layout_char(C).
char_type(vertical, C) :-
	vertical_layout_char(C).

%-------------- Kiiratas -------------------------------------

% print_solutions(+Descr):
% A Descr altal megadott feladat osszes megoldasat
% kiirja az aktualis kimenetre.
print_solutions(Descr) :-
	user:folyo(Descr, River),
	pretty_print(Descr, River),
	fail.
print_solutions(_).

% pretty_print(Descr)
% Kiirja a Descr tablat a megoldas nelkul.
pretty_print(Descr) :-
	pretty_print(Descr, []).

% pretty_print(Descr, River)
% Kiirja a Descr terkep megoldasait River folyo eseten
pretty_print(Descr, River) :-
	Descr = f(R-C,_),
	print_separator(C),
	setof(Point:Value,
		  map_elem(Descr, River, Point, Value),
		  Elems),
	print_lines(1, R, C, Elems).

% map_elem(Descr, River, Point, Value):
% A Descr feladvany River megoldasanak megfelelo terkepen
% a Point koordinataju pontban megjelenitendo a Value ertek.
map_elem(f(_,FieldL), River, Point, Value) :-
	(	[Point|_]=River,
		display_value(forras, Value)
	;	last(River,Point),
		display_value(to, Value)
	;	[_|R0]=River,
		append(R1,[_],R0),
		member(Point, R1),
		Value = @
	;	member(m(R,C,Type), FieldL),
	        non_member(R-C, River),
		Point = R-C,
		display_value(Type, Value)
	).
	
% display_value(Type, Value)
% A Type mezotipus Value ertekkent jelenitendo meg.
display_value(folyo,@).
display_value(forras,'F').
display_value(to,'T').
display_value(ret,'R').
display_value(kikoto(N),N).

% print_lines(R0, R, C, Elems):
% Kiirja az R x C meretu tabla R0-adik sorat,
% az Elems tartalomlistanak megfeleloen.
% A tartalomlista (Sor-Oszlop):Ertek alaku elemekbol all,
% es azt irja elo, hogy a (Sor,Oszlop) pozicion megjelenitendo
% az Ertek konstans.
print_lines(R0, R, _, _) :-
	R0 > R, !.
print_lines(R0, R, C, Elems0) :-
	print_line(1, C, R0, Elems0, Elems1),
	print_separator(C),
	R1 is R0+1,
	print_lines(R1, R, C, Elems1).

% print_line(C0, C, R0, Elems0, Elems):
% Kiirja az R x C meretu tabla R0-adik soranak a C0 oszlopban
% kezdodo reszet az Elems0 tartalomlistanak megfeleloen.
% Az Elems argumentumban visszaadja a tartalomlistanak
% az R0 sor utani reszet.
print_line(C0, C, _, Elems, Elems) :-
	C0 > C, !,
	format('|~n', []).
print_line(C0, C, R0, Elems0, Elems) :-
	field(R0-C0, Value, Elems0, Elems1),
	format('| ~w ', [Value]),
	C1 is C0+1,
	print_line(C1, C, R0, Elems1, Elems).

% field(Point, Value, Elems0, Elems):
% A Point koordinataju pontban, az Elems0 tartalomlista szerint
% megjelenitendo ertek a Value. Elems a Point utani resznek megfelelo
% tartalomlista.
field(Point, Value, [Point:Value|Elems], Elems) :- !.
field(_, ' ', Elems, Elems).
	
% print_separator(N)
% Kiir egy N hossz elvalaszto sort.
print_separator(0) :-
	!, write('+'), nl.
print_separator(N) :-
	write('+---'),
	N1 is N-1,
	print_separator(N1).

%-------------- Seged-eljarasok ------------------------------

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

%-------------- ISO kompatibilitas ---------------------------

% Az alabbi cel az ISO Prolog szabvanynak meg nem megfelelo, de
% DEC10 kompatiblis Prolog rendszerekben valo futtatashoz szukseges.

:- (   predicate_property(get_code(_,_), _) -> true
   ;   assert((get_code(S, C) :- get0(S, C))),
       assert((put_code(C) :- put(C))),
       assert((number_codes(N, C) :- number_chars(N, C)))
   ).
